from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.utils.data
import torch.nn.functional as F
from source import utils

input_dims_per_point = 3
dropout_probability = 0.3
net_size_max = 1024


class STN(nn.Module):
    def __init__(self, num_scales=1, num_points=500, dim=3, sym_op='max'):
        super(STN, self).__init__()

        self.dim = dim
        self.sym_op = sym_op
        self.num_scales = num_scales
        self.num_points = num_points

        self.conv1 = torch.nn.Conv1d(self.dim, 64, 1)
        self.conv2 = torch.nn.Conv1d(64, 128, 1)
        self.conv3 = torch.nn.Conv1d(128, net_size_max, 1)
        self.mp1 = torch.nn.MaxPool1d(num_points)

        self.fc1 = nn.Linear(net_size_max, int(net_size_max / 2))
        self.fc2 = nn.Linear(int(net_size_max / 2), int(net_size_max / 4))
        self.fc3 = nn.Linear(int(net_size_max / 4), self.dim*self.dim)

        self.bn1 = nn.BatchNorm1d(64)
        self.bn2 = nn.BatchNorm1d(128)
        self.bn3 = nn.BatchNorm1d(net_size_max)
        self.bn4 = nn.BatchNorm1d(int(net_size_max / 2))
        self.bn5 = nn.BatchNorm1d(int(net_size_max / 4))

        if self.num_scales > 1:
            self.fc0 = nn.Linear(net_size_max * self.num_scales, net_size_max)
            self.bn0 = nn.BatchNorm1d(net_size_max)

    def forward(self, x):
        batch_size = x.size()[0]
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))

        # symmetric operation over all points
        if self.num_scales == 1:
            x = self.mp1(x)
        else:
            x_scales = x.new_empty(x.size(0), net_size_max * self.num_scales, 1)
            for s in range(self.num_scales):
                x_scales[:, s*net_size_max:(s+1)*net_size_max, :] = \
                    self.mp1(x[:, :, s*self.num_points:(s+1)*self.num_points])
            x = x_scales

        x = x.view(-1, net_size_max*self.num_scales)

        if self.num_scales > 1:
            x = F.relu(self.bn0(self.fc0(x)))

        x = F.relu(self.bn4(self.fc1(x)))
        x = F.relu(self.bn5(self.fc2(x)))
        x = self.fc3(x)

        iden = torch.eye(self.dim, dtype=x.dtype, device=x.device).view(1, self.dim*self.dim).repeat(batch_size, 1)
        x = x + iden
        x = x.view(-1, self.dim, self.dim)
        return x


class QSTN(nn.Module):
    def __init__(self, num_scales=1, num_points=500, dim=3, sym_op='max'):
        super(QSTN, self).__init__()

        self.dim = dim
        self.sym_op = sym_op
        self.num_scales = num_scales
        self.num_points = num_points

        self.conv1 = torch.nn.Conv1d(self.dim, 64, 1)
        self.conv2 = torch.nn.Conv1d(64, 128, 1)
        self.conv3 = torch.nn.Conv1d(128, net_size_max, 1)
        self.mp1 = torch.nn.MaxPool1d(num_points)
        self.fc1 = nn.Linear(net_size_max, int(net_size_max / 2))
        self.fc2 = nn.Linear(int(net_size_max / 2), int(net_size_max / 4))
        self.fc3 = nn.Linear(int(net_size_max / 4), 4)

        self.bn1 = nn.BatchNorm1d(64)
        self.bn2 = nn.BatchNorm1d(128)
        self.bn3 = nn.BatchNorm1d(net_size_max)
        self.bn4 = nn.BatchNorm1d(int(net_size_max / 2))
        self.bn5 = nn.BatchNorm1d(int(net_size_max / 4))

        if self.num_scales > 1:
            self.fc0 = nn.Linear(net_size_max*self.num_scales, net_size_max)
            self.bn0 = nn.BatchNorm1d(net_size_max)

    def forward(self, x):
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))

        # symmetric operation over all points
        if self.num_scales == 1:
            x = self.mp1(x)
        else:
            x_scales = x.new_empty(x.size(0), net_size_max*self.num_scales, 1)
            for s in range(self.num_scales):
                x_scales[:, s*net_size_max:(s+1)*net_size_max, :] = \
                    self.mp1(x[:, :, s*self.num_points:(s+1)*self.num_points])
            x = x_scales

        x = x.view(-1, net_size_max*self.num_scales)

        if self.num_scales > 1:
            x = F.relu(self.bn0(self.fc0(x)))

        x = F.relu(self.bn4(self.fc1(x)))
        x = F.relu(self.bn5(self.fc2(x)))
        x = self.fc3(x)

        # add identity quaternion (so the network can output 0 to leave the point cloud identical)
        iden = x.new_tensor([1, 0, 0, 0])
        x_quat = x + iden

        # convert quaternion to rotation matrix
        x = utils.batch_quat_to_rotmat(x_quat)

        return x, x_quat


class PointNetfeat(nn.Module):
    def __init__(self, num_scales=1, num_points=500, use_point_stn=True, use_feat_stn=True,
                 output_size=100, sym_op='max'):
        super(PointNetfeat, self).__init__()
        self.num_points = num_points
        self.num_scales = num_scales
        self.use_point_stn = use_point_stn
        self.use_feat_stn = use_feat_stn
        self.sym_op = sym_op
        self.output_size = output_size

        if self.use_point_stn:
            # self.stn1 = STN(num_scales=self.num_scales, num_points=num_points,
            #                 dim=input_dims_per_point, sym_op=self.sym_op)
            self.stn1 = QSTN(num_scales=self.num_scales, num_points=num_points,
                             dim=3, sym_op=self.sym_op)

        if self.use_feat_stn:
            self.stn2 = STN(num_scales=self.num_scales, num_points=num_points, dim=64, sym_op=self.sym_op)

        self.conv0a = torch.nn.Conv1d(input_dims_per_point, 64, 1)
        self.conv0b = torch.nn.Conv1d(64, 64, 1)
        self.bn0a = nn.BatchNorm1d(64)
        self.bn0b = nn.BatchNorm1d(64)
        self.conv1 = torch.nn.Conv1d(64, 64, 1)
        self.conv2 = torch.nn.Conv1d(64, 128, 1)
        self.conv3 = torch.nn.Conv1d(128, output_size, 1)
        self.bn1 = nn.BatchNorm1d(64)
        self.bn2 = nn.BatchNorm1d(128)
        self.bn3 = nn.BatchNorm1d(output_size)

        if self.num_scales > 1:
            self.conv4 = torch.nn.Conv1d(output_size, output_size*self.num_scales, 1)
            self.bn4 = nn.BatchNorm1d(output_size*self.num_scales)

        if self.sym_op == 'max':
            self.mp1 = torch.nn.MaxPool1d(num_points)
        elif self.sym_op == 'sum':
            self.mp1 = None
        else:
            raise ValueError('Unsupported symmetric operation: %s' % self.sym_op)

    def forward(self, x):

        # input transform
        if self.use_point_stn:
            # from tuples to list of single points
            x = x.view(x.size(0), input_dims_per_point, -1)
            trans, trans_quat = self.stn1(x[:, :3])  # transform only point data
            x = x.transpose(2, 1)
            # an error here can mean that your input size is wrong (e.g. added normals in the point cloud files)
            x_transformed = torch.bmm(x[:, :, :3], trans)  # transform only point data
            x = torch.cat((x_transformed, x[:, :, 3:]), dim=2)
            x = x.transpose(2, 1)
            x = x.contiguous().view(x.size(0), input_dims_per_point, -1)
        else:
            trans = None
            trans_quat = None

        # mlp (64,64)
        x = F.relu(self.bn0a(self.conv0a(x)))
        x = F.relu(self.bn0b(self.conv0b(x)))

        # feature transform
        if self.use_feat_stn:
            trans2 = self.stn2(x)
            x = x.transpose(2, 1)
            x = torch.bmm(x, trans2)
            x = x.transpose(2, 1)
        else:
            trans2 = None

        # mlp (64,128,output_size)
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = self.bn3(self.conv3(x))

        # mlp (output_size,output_size*num_scales)
        if self.num_scales > 1:
            x = self.bn4(self.conv4(F.relu(x)))

        # symmetric max operation over all points
        if self.num_scales == 1:
            if self.sym_op == 'max':
                x = self.mp1(x)
            elif self.sym_op == 'sum':
                x = torch.sum(x, 2, keepdim=True)
            else:
                raise ValueError('Unsupported symmetric operation: %s' % self.sym_op)

        else:
            x_scales = x.new_empty(x.size(0), self.output_size*self.num_scales**2, 1)
            if self.sym_op == 'max':
                for s in range(self.num_scales):
                    x_scales[:, s*self.num_scales*self.output_size:(s+1)*self.num_scales*self.output_size, :] = \
                        self.mp1(x[:, :, s*self.num_points:(s+1)*self.num_points])
            elif self.sym_op == 'sum':
                for s in range(self.num_scales):
                    x_scales[:, s*self.num_scales*self.output_size:(s+1)*self.num_scales*self.output_size, :] = \
                        torch.sum(x[:, :, s*self.num_points:(s+1)*self.num_points], 2, keepdim=True)
            else:
                raise ValueError('Unsupported symmetric operation: %s' % self.sym_op)
            x = x_scales

        x = x.view(-1, self.output_size * self.num_scales ** 2)

        return x, trans, trans_quat, trans2


class MeshNetGlobalQuery(nn.Module):  # basing on PointNetDenseCls
    def __init__(self, num_points=500, output_dim=3, use_point_stn=True, use_feat_stn=True,
                 sym_op='max', use_query_point=False,
                 sub_sample_size=500, do_augmentation=True, single_transformer=False, shared_transformer=False):
        super(MeshNetGlobalQuery, self).__init__()

        self.num_points = num_points
        self.use_query_point = use_query_point
        self.sub_sample_size = sub_sample_size
        self.num_query_points = int(self.use_query_point)
        self.do_augmentation = do_augmentation
        self.single_transformer = bool(single_transformer)

        if shared_transformer:
            raise ValueError('Shared transformer not supported in this model version.')

        if self.single_transformer:
            self.feat_local_global = PointNetfeat(
                num_points=self.num_points + self.sub_sample_size,
                num_scales=1,
                use_point_stn=use_point_stn,
                use_feat_stn=use_feat_stn,
                # output_size=net_size_max * 2,  # takes similar memory with this
                output_size=net_size_max,
                sym_op=sym_op)
            # self.fc1_local_global = nn.Linear(int(net_size_max) * 2, int(net_size_max))
            self.fc1_local_global = nn.Linear(int(net_size_max), int(net_size_max))
            self.bn1_local_global = nn.BatchNorm1d(int(net_size_max))
        else:
            self.feat_local = PointNetfeat(
                num_points=self.num_points,
                num_scales=1,
                #use_point_stn=use_point_stn,
                use_point_stn=False,
                use_feat_stn=use_feat_stn,
                output_size=net_size_max,
                sym_op=sym_op)
            self.feat_global = PointNetfeat(
                num_points=self.sub_sample_size,
                num_scales=1,
                use_point_stn=use_point_stn,
                #use_point_stn=False,
                use_feat_stn=use_feat_stn,
                output_size=net_size_max,
                sym_op=sym_op)
            self.fc1_local = nn.Linear(int(net_size_max), int(net_size_max / 2))
            self.fc1_global = nn.Linear(int(net_size_max), int(net_size_max / 2))
            self.bn1_local = nn.BatchNorm1d(int(net_size_max / 2))
            self.bn1_global = nn.BatchNorm1d(int(net_size_max / 2))

        self.fc2 = nn.Linear(int(net_size_max / 2) * 2, int(net_size_max / 4))
        self.fc3 = nn.Linear(int(net_size_max / 4), int(net_size_max / 8))
        self.fc4 = nn.Linear(int(net_size_max / 8), output_dim)
        self.bn2 = nn.BatchNorm1d(int(net_size_max / 4))
        self.bn3 = nn.BatchNorm1d(int(net_size_max / 8))

    def forward(self, x):

        patch_features = x['patch_pts_ps'].transpose(1, 2)
        shape_features = x['pts_sub_sample_ms'].transpose(1, 2)
        #patch_query_point = x['imp_surf_query_point_ps'].unsqueeze(2)  # should be zero vector anyway
        shape_query_point = x['imp_surf_query_point_ms'].unsqueeze(2)

        # move global points to query point so that both local and global information are centered at the query point
        shape_features -= shape_query_point.expand(shape_features.shape)

        # again done in data loader on CPU
        # random rotation as data augmentation
        # if self.do_augmentation:
        #     rand_rot_mat = x['rand_rot_mat']
        #     # batched multiplication of vectors (patch points) and matrix (per patch)
        #     patch_features = torch.einsum('jik,jkn->jin', rand_rot_mat, patch_features)
        #     shape_features = torch.einsum('jik,jkn->jin', rand_rot_mat, shape_features)
        #     #patch_query_point = torch.einsum('jik,jkn->jin', rand_rot_mat, patch_query_point)
        #     #shape_query_point = torch.einsum('jik,jkn->jin', rand_rot_mat, shape_query_point)

        #     # random translation as data augmentation
        #     # small translation is sufficient. too much will move the points outside the unit cube
        #     rand_trans = (self.rng.rand(3).astype(np.float32) - 0.5) / 5.0  # [-0.1..+0.1)
        #     # TODO: try to increase the random translation and clip points outside the unit box
        #     pts_sub_sample_ms += np.broadcast_to(rand_trans, pts_sub_sample_ms.shape)
        #     imp_surf_query_point_ms += rand_trans
        #     # don't translate patch points

        ## debug output for a single patch with its sub-sample
        #if True:
        #    from source import utils_eval
        #    out_file = 'debug/train_sample.ply'
        #    utils_eval.visualize_patch(
        #        patch_pts_ps=patch_features[0].detach().cpu().numpy().transpose(),
        #        patch_pts_ms=None,
        #        #query_point_ps=patch_query_point[0].detach().cpu().numpy().transpose().squeeze(),
        #        query_point_ps=torch.zeros(3),
        #        pts_sub_sample_ms=shape_features[0].detach().cpu().numpy().transpose(),
        #        #query_point_ms=shape_query_point[0].detach().cpu().numpy().transpose().squeeze(),
        #        query_point_ms=torch.zeros(3),
        #        file_path=out_file)
        #    print('wrote training sample to {}'.format(out_file))

        if self.single_transformer:
            local_global_features = torch.cat((patch_features,  shape_features), dim=2)
            local_global_features_transformed, _, _, _ = self.feat_local_global(local_global_features)
            patch_features = F.relu(self.bn1_local_global(self.fc1_local_global(local_global_features_transformed)))
        else:
            shape_features, trans_global_pts, trans_global_pts_quat, trans_global_feat = \
                self.feat_global(shape_features)
            shape_features = F.relu(self.bn1_global(self.fc1_global(shape_features)))

            # rotate patch-space points like the subsample
            patch_features = torch.bmm(patch_features.transpose(2, 1), trans_global_pts).transpose(2, 1)

            patch_features, trans_local_pts, trans_local_pts_quat, trans_local_feat = \
                self.feat_local(patch_features)
            patch_features = F.relu(self.bn1_local(self.fc1_local(patch_features)))

            # rotate query points like the patch
            patch_features = torch.cat((patch_features,  shape_features), dim=1)

        patch_features = F.relu(self.bn2(self.fc2(patch_features)))
        patch_features = F.relu(self.bn3(self.fc3(patch_features)))
        patch_features = self.fc4(patch_features)

        return patch_features


class MeshNetLocalQuery(nn.Module):  # basing on PointNetDenseCls
    def __init__(self, num_points=500, output_dim=3, use_point_stn=True, use_feat_stn=True,
                 sym_op='max', use_query_point=False, sub_sample_size=500):
        super(MeshNetLocalQuery, self).__init__()

        self.num_points = num_points
        self.use_query_point = use_query_point
        self.sub_sample_size = sub_sample_size
        self.num_query_points = int(self.use_query_point)

        self.feat = PointNetfeat(
            num_points=num_points,
            num_scales=1,
            use_point_stn=use_point_stn,
            use_feat_stn=use_feat_stn,
            sym_op=sym_op)
        self.fc1 = nn.Linear(int(net_size_max), int(net_size_max / 2))
        self.fc2 = nn.Linear(int(net_size_max / 2) + self.num_query_points * 3, int(net_size_max / 4))
        self.fc3 = nn.Linear(int(net_size_max / 4), output_dim)
        self.bn1 = nn.BatchNorm1d(int(net_size_max / 2))
        self.bn2 = nn.BatchNorm1d(int(net_size_max / 4))

    def forward(self, x):

        patch_features = x[:, :, :self.num_points]
        patch_query_point = x[:, :, self.num_points:self.num_points+self.num_query_points]

        patch_features, trans, trans2, pointfvals = self.feat(patch_features)
        patch_features = F.relu(self.bn1(self.fc1(patch_features)))
        #patch_features = self.do1(patch_features)

        if self.use_query_point:
            patch_query_point_transformed = torch.bmm(patch_query_point.transpose(2, 1), trans).transpose(2, 1)
            patch_features = torch.cat((patch_features, patch_query_point_transformed.squeeze(dim=2)), dim=1)

        patch_features = F.relu(self.bn2(self.fc2(patch_features)))
        #patch_features = self.do2(patch_features)
        patch_features = self.fc3(patch_features)

        return patch_features, trans, trans2, pointfvals


class MeshNetLocal(nn.Module):  # basing on PointNetDenseCls
    def __init__(self, num_points=500, output_dim=3, use_point_stn=True, use_feat_stn=True,
                 sym_op='max', use_query_point=False, sub_sample_size=None):
        super(MeshNetLocal, self).__init__()

        if use_query_point:  # query point will be added to inputs
            num_points += 1

        self.num_points = num_points

        self.feat = PointNetfeat(
            num_points=num_points,
            num_scales=1,
            use_point_stn=use_point_stn,
            use_feat_stn=use_feat_stn,
            sym_op=sym_op)
        self.fc1 = nn.Linear(int(net_size_max), int(net_size_max / 2))
        self.fc2 = nn.Linear(int(net_size_max / 2), int(net_size_max / 4))
        self.fc3 = nn.Linear(int(net_size_max / 4), output_dim)
        self.bn1 = nn.BatchNorm1d(int(net_size_max / 2))
        self.bn2 = nn.BatchNorm1d(int(net_size_max / 4))
        self.do1 = nn.Dropout(p=dropout_probability)
        self.do2 = nn.Dropout(p=dropout_probability)

    def forward(self, x):
        x, trans, trans2, pointfvals = self.feat(x)
        x = F.relu(self.bn1(self.fc1(x)))
        x = self.do1(x)
        x = F.relu(self.bn2(self.fc2(x)))
        x = self.do2(x)
        x = self.fc3(x)

        return x, trans, trans2, pointfvals


class MSMeshNet(nn.Module):
    def __init__(self, num_scales=2, num_points=500, output_dim=3, use_point_stn=True,
                 use_feat_stn=True, sym_op='max'):
        super(MSMeshNet, self).__init__()
        self.num_points = num_points

        self.feat = PointNetfeat(
            num_points=num_points,
            num_scales=num_scales,
            use_point_stn=use_point_stn,
            use_feat_stn=use_feat_stn,
            sym_op=sym_op)
        self.fc0 = nn.Linear(net_size_max*num_scales**2, net_size_max)
        self.fc1 = nn.Linear(net_size_max, int(net_size_max / 2))
        self.fc2 = nn.Linear(int(net_size_max / 2), int(net_size_max / 4))
        self.fc3 = nn.Linear(int(net_size_max / 4), output_dim)
        self.bn0 = nn.BatchNorm1d(net_size_max)
        self.bn1 = nn.BatchNorm1d(int(net_size_max / 2))
        self.bn2 = nn.BatchNorm1d(int(net_size_max / 4))
        self.do1 = nn.Dropout(p=dropout_probability)
        self.do2 = nn.Dropout(p=dropout_probability)

    def forward(self, x):
        x, trans, trans2, pointfvals = self.feat(x)
        x = F.relu(self.bn0(self.fc0(x)))
        x = F.relu(self.bn1(self.fc1(x)))
        x = self.do1(x)
        x = F.relu(self.bn2(self.fc2(x)))
        x = self.do2(x)
        x = self.fc3(x)

        return x, trans, trans2, pointfvals
