import numpy as np
import os
import scipy.sparse as sparse

from source import utils_mp
from source import utils_files


def compare_predictions_multi_class(ground_truth, predicted, prediction_name):
    classes = np.unique(ground_truth)

    results = []
    for class_val in classes:
        gt_class = ground_truth == class_val
        pred_class = predicted == class_val
        results.append(compare_predictions_binary_dense(gt_class, pred_class, str(class_val) + '_' + prediction_name))

    return results


def calc_accuracy(num_true, num_predictions):
    if num_predictions == 0:
        return float('NaN')
    else:
        return num_true / num_predictions


def calc_precision(num_true_pos, num_false_pos):
    if isinstance(num_true_pos, (int, float)) and isinstance(num_false_pos, (int, float)) and \
            num_true_pos + num_false_pos == 0:
        return float('NaN')
    else:
        return num_true_pos / (num_true_pos + num_false_pos)


def calc_recall(num_true_pos, num_false_neg):
    if isinstance(num_true_pos, (int, float)) and isinstance(num_false_neg, (int, float)) and \
            num_true_pos + num_false_neg == 0:
        return float('NaN')
    else:
        return num_true_pos / (num_true_pos + num_false_neg)


def calc_f1(precision, recall):
    if isinstance(precision, (int, float)) and isinstance(recall, (int, float)) and \
            precision + recall == 0:
        return float('NaN')
    else:
        return 2.0 * (precision * recall) / (precision + recall)


def compare_predictions_binary_dense(ground_truth, predicted, prediction_name):
    """

    :param ground_truth:
    :param predicted:
    :param prediction_name:
    :return: res_dict, prec_per_patch
    """

    # from sklearn.metrics import precision_recall_fscore_support

    if ground_truth.shape != predicted.shape:
        raise ValueError('The ground truth matrix and the predicted matrix have different sizes!')
        
    if not isinstance(ground_truth, np.ndarray) and not isinstance(predicted, np.ndarray):
        raise ValueError('Both matrices must be dense of type np.ndarray!')

    ground_truth_int = (ground_truth > 0.0).astype(int, copy=False)
    predicted_int = (predicted > 0.0).astype(int, copy=False)
    res_dict = dict()
    res_dict['comp_name'] = prediction_name
    
    res_dict["predictions"] = float(np.prod(ground_truth_int.shape))
    res_dict["pred_gt"] = float(np.prod(ground_truth_int.shape))
    res_dict["positives"] = float(np.count_nonzero(predicted_int))
    res_dict["pos_gt"] = float(np.count_nonzero(ground_truth_int))
    res_dict["true_neg"] = res_dict["predictions"] - float(np.count_nonzero(predicted_int + ground_truth_int))
    res_dict["negatives"] = res_dict["predictions"] - res_dict["positives"]
    res_dict["neg_gt"] = res_dict["pred_gt"] - res_dict["pos_gt"]
    true_pos = ((predicted_int + ground_truth_int) == 2).sum(axis=1, dtype=np.float)
    res_dict["true_pos"] = float(true_pos.sum())
    res_dict["true"] = res_dict["true_pos"] + res_dict["true_neg"]
    false_pos = ((predicted_int * 2 + ground_truth_int) == 2).sum(axis=1, dtype=np.float)
    res_dict["false_pos"] = float(false_pos.sum())
    false_neg = ((predicted_int + 2 * ground_truth_int) == 2).sum(axis=1, dtype=np.float)
    res_dict["false_neg"] = float(false_neg.sum())
    res_dict["false"] = res_dict["false_pos"] + res_dict["false_neg"]
    res_dict["accuracy"] = calc_accuracy(res_dict["true"], res_dict["predictions"])
    res_dict["precision"] = calc_precision(res_dict["true_pos"], res_dict["false_pos"])
    res_dict["recall"] = calc_recall(res_dict["true_pos"], res_dict["false_neg"])
    res_dict["f1_score"] = calc_f1(res_dict["precision"], res_dict["recall"])
    # (prec, recall, f1, _) = \
    #     precision_recall_fscore_support(ground_truth_int.flatten(), predicted_int.flatten(), average='binary')
    # res_dict["precision"] = prec.item()
    # res_dict["recall"] = recall.item()
    # res_dict["f1_score"] = f1.item()

    prec_per_patch = calc_precision(true_pos, false_pos)  # true_pos / (true_pos + false_pos)
    recall_per_patch = calc_recall(true_pos, false_neg)
    f1_per_patch = calc_f1(prec_per_patch, recall_per_patch)

    return res_dict, prec_per_patch, recall_per_patch, f1_per_patch


def compare_predictions_binary_tensors(ground_truth, predicted, prediction_name):
    """

    :param ground_truth:
    :param predicted:
    :param prediction_name:
    :return: res_dict, prec_per_patch
    """

    import torch

    if ground_truth.shape != predicted.shape:
        raise ValueError('The ground truth matrix and the predicted matrix have different sizes!')

    if not isinstance(ground_truth, torch.Tensor) and not isinstance(predicted, torch.Tensor):
        raise ValueError('Both matrices must be dense of type torch.tensor!')

    ground_truth_int = (ground_truth > 0.0).to(dtype=torch.int32)
    predicted_int = (predicted > 0.0).to(dtype=torch.int32)
    res_dict = dict()
    res_dict['comp_name'] = prediction_name

    res_dict["predictions"] = float(torch.numel(ground_truth_int))
    res_dict["pred_gt"] = float(torch.numel(ground_truth_int))
    res_dict["positives"] = float(torch.nonzero(predicted_int).shape[0])
    res_dict["pos_gt"] = float(torch.nonzero(ground_truth_int).shape[0])
    res_dict["true_neg"] = res_dict["predictions"] - float(torch.nonzero(predicted_int + ground_truth_int).shape[0])
    res_dict["negatives"] = res_dict["predictions"] - res_dict["positives"]
    res_dict["neg_gt"] = res_dict["pred_gt"] - res_dict["pos_gt"]
    true_pos = ((predicted_int + ground_truth_int) == 2).sum().to(dtype=torch.float32)
    res_dict["true_pos"] = float(true_pos.sum())
    res_dict["true"] = res_dict["true_pos"] + res_dict["true_neg"]
    false_pos = ((predicted_int * 2 + ground_truth_int) == 2).sum().to(dtype=torch.float32)
    res_dict["false_pos"] = float(false_pos.sum())
    false_neg = ((predicted_int + 2 * ground_truth_int) == 2).sum().to(dtype=torch.float32)
    res_dict["false_neg"] = float(false_neg.sum())
    res_dict["false"] = res_dict["false_pos"] + res_dict["false_neg"]
    res_dict["accuracy"] = calc_accuracy(res_dict["true"], res_dict["predictions"])
    res_dict["precision"] = calc_precision(res_dict["true_pos"], res_dict["false_pos"])
    res_dict["recall"] = calc_recall(res_dict["true_pos"], res_dict["false_neg"])
    res_dict["f1_score"] = calc_f1(res_dict["precision"], res_dict["recall"])

    return res_dict


def sparse_dist_matrix_to_dense(sparse_mat, points, points_per_patch):
    """
    Resolve patches and turn the information into a dense matrix.
    :param sparse_mat:
    :param points:
    :param points_per_patch:
    :return:
    """

    if not isinstance(sparse_mat, sparse.spmatrix):
        raise ValueError('The matrix must be a sparse matrix of type scipy.sparse.spmatrix!')

    import sys
    import scipy.spatial as spatial
    # otherwise KDTree construction may run out of recursions
    leaf_size = 100
    sys.setrecursionlimit(int(max(1000, round(points.shape[0] / leaf_size))))
    kdtree = spatial.cKDTree(points, leaf_size)

    query_points = points[range(sparse_mat.shape[0]), :]
    patch_pts_dists, patch_pts_ids = kdtree.query(x=query_points, k=points_per_patch, n_jobs=-1)
    patch_ids = np.expand_dims(np.arange(sparse_mat.shape[0]), axis=1).repeat(points_per_patch, axis=1)
    sparse_nn = sparse_mat[patch_ids, patch_pts_ids]
    dense_nn = sparse_nn.A.squeeze()

    return dense_nn


def compare_predictions_binary_sparse(
        ground_truth,
        predicted,
        prediction_name,
        points,
        points_per_patch):
    """
    Mean scores over all patches
    :param ground_truth:
    :param predicted:
    :param prediction_name:
    :param points:
    :param points_per_patch:
    :param pred_is_geo_dist:
    :return: res_dict, prec_per_patch
    """

    if ground_truth.shape != predicted.shape:
        raise ValueError('The ground truth matrix and the predicted matrix have different sizes!')

    if not isinstance(ground_truth, sparse.spmatrix) and not isinstance(predicted, sparse.spmatrix):
        raise ValueError('Both matrices must be sparse of type scipy.sparse.spmatrix!')

    gt_geo_dense = sparse_dist_matrix_to_dense(ground_truth, points, points_per_patch)
    pred_geo_dense = sparse_dist_matrix_to_dense(predicted, points, points_per_patch)

    gt_geo_conn_dense = (gt_geo_dense >= 0.0).astype(int, copy=False)  # >= 0 because disconnected is -1
    pred_geo_conn_dense = (pred_geo_dense > 0.0).astype(int, copy=False)  # > 0 because disconnected is 0

    res_dict, prec_per_patch, recall_per_patch, f1_per_patch = \
        compare_predictions_binary_dense(gt_geo_conn_dense, pred_geo_conn_dense, prediction_name)

    return res_dict, prec_per_patch, recall_per_patch, f1_per_patch


def eval_predictions(pred_path, gt_path, report_file=None, unsigned=False):
    files = [f for f in os.listdir(pred_path) if os.path.isfile(os.path.join(pred_path, f)) and f[-4:] == '.npy']

    results = []
    for f in files:
        gt_off_path = os.path.join(gt_path, f[:-8] + '.ply.npy')
        rec_off_path = os.path.join(pred_path, f)

        mat_gt = np.load(gt_off_path)
        mat_rec = np.load(rec_off_path)

        if unsigned:
            mat_gt = np.abs(mat_gt)
            mat_rec = np.abs(mat_rec)

        gt_or_pred_nz = ((mat_rec != 0.0) + (mat_gt != 0.0)) > 0
        l2 = (mat_rec - mat_gt)
        l2_sq = l2 * l2
        mse = l2_sq[gt_or_pred_nz].mean()

        mat_gt_mean = mat_gt.mean()
        mat_rec_mean = mat_rec.mean()
        mat_gt_var = (mat_gt * mat_gt).mean() - mat_gt_mean * mat_gt_mean
        mat_rec_var = (mat_rec * mat_rec).mean() - mat_rec_mean * mat_rec_mean

        res_dict = {
            'file': f,
            'mse': mse,
            'mean_gt': mat_gt_mean,
            'mean_pred': mat_rec_mean,
            'var_gt': mat_gt_var,
            'var_pred': mat_rec_var,
        }
        results.append(res_dict)

    print('compare_prediction: {} vs {}\n'.format(gt_path, pred_path))
    lines = print_list_of_dicts(results, ['file', 'mse', 'mean_gt', 'mean_pred', 'var_gt', 'var_pred'], mode='csv')

    if report_file is not None:
        utils_files.make_dir_for_file(report_file)
        with open(report_file, 'w') as the_file:
            for l in lines:
                the_file.write(l + '\n')


def print_list_of_dicts(comp_res, keys_to_print=None, mode='latex'):

    if len(comp_res) == 0:
        return 'WARNING: comp_res is empty'

    if keys_to_print is None or len(keys_to_print) == 0:
        keys_to_print = comp_res[0].keys()

    def get_separator(i, length):
        if mode == 'latex':
            if i < length - 1:
                return ' & '
            else:
                return ' \\\\'
        elif mode == 'csv':
            return ','
            
    # key per line, mesh per collumn
    #for key in keys_to_print:
    #    line = key + ' && '
    #    for i, d in enumerate(comp_res):
    #        if isinstance(d[key], str):
    #            line += d[key] + get_separator(i, len(keys_to_print))
    #        else:
    #            line += '{0:.3f}'.format(d[key]) + get_separator(i, len(keys_to_print))
    #    print(line)
    
    # mesh per line, key per column
    lines = []
    # contents
    for d in comp_res:
        line = ''
        for i, key in enumerate(keys_to_print):
            if isinstance(d[key], str):
                line += d[key][:10].replace('_', ' ').rjust(max(10, len(key))) + get_separator(i, len(keys_to_print))
            else:
                line += '{0:.5f}'.format(d[key]).rjust(max(10, len(key))) + get_separator(i, len(keys_to_print))
        lines.append(line)
    
    lines.sort()
    
    # header
    line = ''
    for i, key in enumerate(keys_to_print):
        line += key.replace('_', ' ').rjust(10) + get_separator(i, len(keys_to_print))
    lines.insert(0, line)

    for l in lines:
        print(l)

    return lines


def visualize_patch(patch_pts_ps, patch_pts_ms, query_point_ps, pts_sub_sample_ms, query_point_ms,
                    file_path='debug/patch.ply'):

    from source import point_cloud

    def filter_padding(patch_pts, query_point):
        query_point_repeated = np.repeat(np.expand_dims(np.array(query_point), axis=0), patch_pts.shape[0], axis=0)
        same_points = patch_pts == query_point_repeated
        non_padding_point_ids = np.sum(same_points, axis=1) != 3
        return patch_pts[non_padding_point_ids]

    patch_pts_ps = filter_padding(patch_pts_ps, query_point_ps)
    if patch_pts_ms is not None:
        patch_pts_ms = filter_padding(patch_pts_ms, query_point_ms)

    query_point_ps = np.expand_dims(query_point_ps, axis=0) \
        if len(query_point_ps.shape) < 2 else query_point_ps
    query_point_ms = np.expand_dims(query_point_ms, axis=0) \
        if len(query_point_ms.shape) < 2 else query_point_ms

    pts = np.concatenate((patch_pts_ps, query_point_ps, pts_sub_sample_ms, query_point_ms), axis=0)
    if patch_pts_ms is not None:
        pts = np.concatenate((pts, patch_pts_ms), axis=0)

    def repeat_color_for_points(color, points):
        return np.repeat(np.expand_dims(np.array(color), axis=0), points.shape[0], axis=0)

    colors_patch_pts_ps = repeat_color_for_points([0.0, 0.0, 1.0], patch_pts_ps)
    if patch_pts_ms is not None:
        colors_patch_pts_ms = repeat_color_for_points([1.0, 0.0, 0.0], patch_pts_ms)
    colors_query_point_ps = repeat_color_for_points([1.0, 1.0, 0.0], query_point_ps)
    colors_pts_sub_sample_ms = repeat_color_for_points([0.0, 1.0, 0.0], pts_sub_sample_ms)
    colors_query_point_ms = repeat_color_for_points([1.0, 0.0, 1.0], query_point_ms)
    colors = np.concatenate((colors_patch_pts_ps, colors_query_point_ps, colors_pts_sub_sample_ms,
                             colors_query_point_ms), axis=0)
    if patch_pts_ms is not None:
        colors = np.concatenate(
            (colors, colors_patch_pts_ms), axis=0)

    point_cloud.write_ply(file_path=file_path, points=pts, colors=colors)


def _chamfer_distance_single_file(file_in, file_ref, samples_per_model, num_processes=1):
    # http://graphics.stanford.edu/courses/cs468-17-spring/LectureSlides/L14%20-%203d%20deep%20learning%20on%20point%20cloud%20representation%20(analysis).pdf

    import trimesh
    import trimesh.sample
    import sys
    import scipy.spatial as spatial

    def sample_mesh(mesh_file, num_samples):
        try:
            mesh = trimesh.load(mesh_file)
        except:
            return np.zeros((0, 3))
        samples, face_indices = trimesh.sample.sample_surface_even(mesh, num_samples)
        return samples

    new_mesh_samples = sample_mesh(file_in, samples_per_model)
    ref_mesh_samples = sample_mesh(file_ref, samples_per_model)

    if new_mesh_samples.shape[0] == 0 or ref_mesh_samples.shape[0] == 0:
        return file_in, file_ref, -1.0

    leaf_size = 100
    sys.setrecursionlimit(int(max(1000, round(new_mesh_samples.shape[0] / leaf_size))))
    kdtree_new_mesh_samples = spatial.cKDTree(new_mesh_samples, leaf_size)
    kdtree_ref_mesh_samples = spatial.cKDTree(ref_mesh_samples, leaf_size)

    ref_new_dist, corr_new_ids = kdtree_new_mesh_samples.query(ref_mesh_samples, 1, n_jobs=num_processes)
    new_ref_dist, corr_ref_ids = kdtree_ref_mesh_samples.query(new_mesh_samples, 1, n_jobs=num_processes)

    ref_new_dist_sum = np.sum(ref_new_dist)
    new_ref_dist_sum = np.sum(new_ref_dist)
    chamfer_dist = ref_new_dist_sum + new_ref_dist_sum

    return file_in, file_ref, chamfer_dist


def _hausdorff_distance_directed_single_file(file_in, file_ref, samples_per_model):
    import scipy.spatial as spatial
    import trimesh
    import trimesh.sample

    def sample_mesh(mesh_file, num_samples):
        try:
            mesh = trimesh.load(mesh_file)
        except:
            return np.zeros((0, 3))
        samples, face_indices = trimesh.sample.sample_surface_even(mesh, num_samples)
        return samples

    new_mesh_samples = sample_mesh(file_in, samples_per_model)
    ref_mesh_samples = sample_mesh(file_ref, samples_per_model)

    if new_mesh_samples.shape[0] == 0 or ref_mesh_samples.shape[0] == 0:
        return file_in, file_ref, -1.0

    dist, _, _ = spatial.distance.directed_hausdorff(new_mesh_samples, ref_mesh_samples)
    return file_in, file_ref, dist


def _hausdorff_distance_single_file(file_in, file_ref, samples_per_model):
    import scipy.spatial as spatial
    import trimesh
    import trimesh.sample

    def sample_mesh(mesh_file, num_samples):
        try:
            mesh = trimesh.load(mesh_file)
        except:
            return np.zeros((0, 3))
        samples, face_indices = trimesh.sample.sample_surface_even(mesh, num_samples)
        return samples

    new_mesh_samples = sample_mesh(file_in, samples_per_model)
    ref_mesh_samples = sample_mesh(file_ref, samples_per_model)

    if new_mesh_samples.shape[0] == 0 or ref_mesh_samples.shape[0] == 0:
        return file_in, file_ref, -1.0, -1.0, -1.0

    dist_new_ref, _, _ = spatial.distance.directed_hausdorff(new_mesh_samples, ref_mesh_samples)
    dist_ref_new, _, _ = spatial.distance.directed_hausdorff(ref_mesh_samples, new_mesh_samples)
    dist = max(dist_new_ref, dist_ref_new)
    return file_in, file_ref, dist_new_ref, dist_ref_new, dist


def mesh_comparison(new_meshes_dir_abs, ref_meshes_dir_abs,
                    num_processes, report_name, samples_per_model=10000, dataset_file_abs=None):
    if not os.path.isdir(new_meshes_dir_abs):
        print('Warning: dir to check doesn\'t exist'.format(new_meshes_dir_abs))
        return

    new_mesh_files = [f for f in os.listdir(new_meshes_dir_abs)
                      if os.path.isfile(os.path.join(new_meshes_dir_abs, f))]
    ref_mesh_files = [f for f in os.listdir(ref_meshes_dir_abs)
                      if os.path.isfile(os.path.join(ref_meshes_dir_abs, f))]

    if dataset_file_abs is None:
        mesh_files_to_compare_set = set(ref_mesh_files)  # set for efficient search
    else:
        if not os.path.isfile(dataset_file_abs):
            raise ValueError('File does not exist: {}'.format(dataset_file_abs))
        with open(dataset_file_abs) as f:
            mesh_files_to_compare_set = f.readlines()
            mesh_files_to_compare_set = [f.replace('\n', '') + '.ply' for f in mesh_files_to_compare_set]
            mesh_files_to_compare_set = [f.split('.')[0] for f in mesh_files_to_compare_set]
            mesh_files_to_compare_set = set(mesh_files_to_compare_set)

    # skip if everything is unchanged
    new_mesh_files_abs = [os.path.join(new_meshes_dir_abs, f) for f in new_mesh_files]
    ref_mesh_files_abs = [os.path.join(ref_meshes_dir_abs, f) for f in ref_mesh_files]
    #if not utils_files.call_necessary(new_mesh_files_abs + ref_mesh_files_abs, report_name):
    #    return

    def ref_mesh_for_new_mesh(new_mesh_file: str, all_ref_meshes: list) -> list:
        stem_new_mesh_file = new_mesh_file.split('.')[0]
        ref_files = list(set([f for f in all_ref_meshes if f.split('.')[0] == stem_new_mesh_file]))
        return ref_files

    call_params = []
    for fi, new_mesh_file in enumerate(new_mesh_files):
        if new_mesh_file.split('.')[0] in mesh_files_to_compare_set:
            new_mesh_file_abs = os.path.join(new_meshes_dir_abs, new_mesh_file)
            ref_mesh_files_matching = ref_mesh_for_new_mesh(new_mesh_file, ref_mesh_files)
            if len(ref_mesh_files_matching) > 0:
                ref_mesh_file_abs = os.path.join(ref_meshes_dir_abs, ref_mesh_files_matching[0])
                call_params.append((new_mesh_file_abs, ref_mesh_file_abs, samples_per_model))
    if len(call_params) == 0:
        raise ValueError('Results are empty!')
    results_hausdorff = utils_mp.start_process_pool(_hausdorff_distance_single_file, call_params, num_processes)
    results = [(r[0], r[1], str(r[2]), str(r[3]), str(r[4])) for r in results_hausdorff]

    call_params = []
    for fi, new_mesh_file in enumerate(new_mesh_files):
        if new_mesh_file.split('.')[0] in mesh_files_to_compare_set:
            new_mesh_file_abs = os.path.join(new_meshes_dir_abs, new_mesh_file)
            ref_mesh_files_matching = ref_mesh_for_new_mesh(new_mesh_file, ref_mesh_files)
            if len(ref_mesh_files_matching) > 0:
                ref_mesh_file_abs = os.path.join(ref_meshes_dir_abs, ref_mesh_files_matching[0])
                call_params.append((new_mesh_file_abs, ref_mesh_file_abs, samples_per_model))
    results_chamfer = utils_mp.start_process_pool(_chamfer_distance_single_file, call_params, num_processes)
    results = [r + (str(results_chamfer[ri][2]),) for ri, r in enumerate(results)]

    # no reference but reconstruction
    for fi, new_mesh_file in enumerate(new_mesh_files):
        if new_mesh_file.split('.')[0] not in mesh_files_to_compare_set:
            if dataset_file_abs is None:
                new_mesh_file_abs = os.path.join(new_meshes_dir_abs, new_mesh_file)
                ref_mesh_files_matching = ref_mesh_for_new_mesh(new_mesh_file, ref_mesh_files)
                if len(ref_mesh_files_matching) > 0:
                    reference_mesh_file_abs = os.path.join(ref_meshes_dir_abs, ref_mesh_files_matching[0])
                    results.append((new_mesh_file_abs, reference_mesh_file_abs, str(-2), str(-2), str(-2), str(-2)))
        else:
            mesh_files_to_compare_set.remove(new_mesh_file.split('.')[0])

    # no reconstruction but reference
    for ref_without_new_mesh in mesh_files_to_compare_set:
        new_mesh_file_abs = os.path.join(new_meshes_dir_abs, ref_without_new_mesh)
        reference_mesh_file_abs = os.path.join(ref_meshes_dir_abs, ref_without_new_mesh)
        results.append((new_mesh_file_abs, reference_mesh_file_abs, str(-1), str(-1), str(-1), str(-1)))

    # sort by file name
    results = sorted(results, key=lambda x: x[0])

    utils_files.make_dir_for_file(report_name)
    csv_lines = ['in mesh,ref mesh,Hausdorff dist new-ref,Hausdorff dist ref-new,Hausdorff dist,'
                 'Chamfer dist(-1: no input; -2: no reference)']
    csv_lines += [','.join(item) for item in results]
    #csv_lines += ['=AVERAGE(E2:E41)']
    csv_lines_str = '\n'.join(csv_lines)
    with open(report_name, "w") as text_file:
        text_file.write(csv_lines_str)


def visualize_distances(in_file_rec_meshes: str, in_file_gt_meshes: str, out_file_vis_meshes: str,
                        normalize_to: float):

    import numpy as np
    import trimesh.proximity

    # for hsv color map
    # from skimage.color import hsv2rgb

    parula_cm = np.array([
        [0.2422, 0.1504, 0.6603],
        [0.2444, 0.1534, 0.6728],
        [0.2464, 0.1569, 0.6847],
        [0.2484, 0.1607, 0.6961],
        [0.2503, 0.1648, 0.7071],
        [0.2522, 0.1689, 0.7179],
        [0.254, 0.1732, 0.7286],
        [0.2558, 0.1773, 0.7393],
        [0.2576, 0.1814, 0.7501],
        [0.2594, 0.1854, 0.761],
        [0.2611, 0.1893, 0.7719],
        [0.2628, 0.1932, 0.7828],
        [0.2645, 0.1972, 0.7937],
        [0.2661, 0.2011, 0.8043],
        [0.2676, 0.2052, 0.8148],
        [0.2691, 0.2094, 0.8249],
        [0.2704, 0.2138, 0.8346],
        [0.2717, 0.2184, 0.8439],
        [0.2729, 0.2231, 0.8528],
        [0.274, 0.228, 0.8612],
        [0.2749, 0.233, 0.8692],
        [0.2758, 0.2382, 0.8767],
        [0.2766, 0.2435, 0.884],
        [0.2774, 0.2489, 0.8908],
        [0.2781, 0.2543, 0.8973],
        [0.2788, 0.2598, 0.9035],
        [0.2794, 0.2653, 0.9094],
        [0.2798, 0.2708, 0.915],
        [0.2802, 0.2764, 0.9204],
        [0.2806, 0.2819, 0.9255],
        [0.2809, 0.2875, 0.9305],
        [0.2811, 0.293, 0.9352],
        [0.2813, 0.2985, 0.9397],
        [0.2814, 0.304, 0.9441],
        [0.2814, 0.3095, 0.9483],
        [0.2813, 0.315, 0.9524],
        [0.2811, 0.3204, 0.9563],
        [0.2809, 0.3259, 0.96],
        [0.2807, 0.3313, 0.9636],
        [0.2803, 0.3367, 0.967],
        [0.2798, 0.3421, 0.9702],
        [0.2791, 0.3475, 0.9733],
        [0.2784, 0.3529, 0.9763],
        [0.2776, 0.3583, 0.9791],
        [0.2766, 0.3638, 0.9817],
        [0.2754, 0.3693, 0.984],
        [0.2741, 0.3748, 0.9862],
        [0.2726, 0.3804, 0.9881],
        [0.271, 0.386, 0.9898],
        [0.2691, 0.3916, 0.9912],
        [0.267, 0.3973, 0.9924],
        [0.2647, 0.403, 0.9935],
        [0.2621, 0.4088, 0.9946],
        [0.2591, 0.4145, 0.9955],
        [0.2556, 0.4203, 0.9965],
        [0.2517, 0.4261, 0.9974],
        [0.2473, 0.4319, 0.9983],
        [0.2424, 0.4378, 0.9991],
        [0.2369, 0.4437, 0.9996],
        [0.2311, 0.4497, 0.9995],
        [0.225, 0.4559, 0.9985],
        [0.2189, 0.462, 0.9968],
        [0.2128, 0.4682, 0.9948],
        [0.2066, 0.4743, 0.9926],
        [0.2006, 0.4803, 0.9906],
        [0.195, 0.4861, 0.9887],
        [0.1903, 0.4919, 0.9867],
        [0.1869, 0.4975, 0.9844],
        [0.1847, 0.503, 0.9819],
        [0.1831, 0.5084, 0.9793],
        [0.1818, 0.5138, 0.9766],
        [0.1806, 0.5191, 0.9738],
        [0.1795, 0.5244, 0.9709],
        [0.1785, 0.5296, 0.9677],
        [0.1778, 0.5349, 0.9641],
        [0.1773, 0.5401, 0.9602],
        [0.1768, 0.5452, 0.956],
        [0.1764, 0.5504, 0.9516],
        [0.1755, 0.5554, 0.9473],
        [0.174, 0.5605, 0.9432],
        [0.1716, 0.5655, 0.9393],
        [0.1686, 0.5705, 0.9357],
        [0.1649, 0.5755, 0.9323],
        [0.161, 0.5805, 0.9289],
        [0.1573, 0.5854, 0.9254],
        [0.154, 0.5902, 0.9218],
        [0.1513, 0.595, 0.9182],
        [0.1492, 0.5997, 0.9147],
        [0.1475, 0.6043, 0.9113],
        [0.1461, 0.6089, 0.908],
        [0.1446, 0.6135, 0.905],
        [0.1429, 0.618, 0.9022],
        [0.1408, 0.6226, 0.8998],
        [0.1383, 0.6272, 0.8975],
        [0.1354, 0.6317, 0.8953],
        [0.1321, 0.6363, 0.8932],
        [0.1288, 0.6408, 0.891],
        [0.1253, 0.6453, 0.8887],
        [0.1219, 0.6497, 0.8862],
        [0.1185, 0.6541, 0.8834],
        [0.1152, 0.6584, 0.8804],
        [0.1119, 0.6627, 0.877],
        [0.1085, 0.6669, 0.8734],
        [0.1048, 0.671, 0.8695],
        [0.1009, 0.675, 0.8653],
        [0.0964, 0.6789, 0.8609],
        [0.0914, 0.6828, 0.8562],
        [0.0855, 0.6865, 0.8513],
        [0.0789, 0.6902, 0.8462],
        [0.0713, 0.6938, 0.8409],
        [0.0628, 0.6972, 0.8355],
        [0.0535, 0.7006, 0.8299],
        [0.0433, 0.7039, 0.8242],
        [0.0328, 0.7071, 0.8183],
        [0.0234, 0.7103, 0.8124],
        [0.0155, 0.7133, 0.8064],
        [0.0091, 0.7163, 0.8003],
        [0.0046, 0.7192, 0.7941],
        [0.0019, 0.722, 0.7878],
        [0.0009, 0.7248, 0.7815],
        [0.0018, 0.7275, 0.7752],
        [0.0046, 0.7301, 0.7688],
        [0.0094, 0.7327, 0.7623],
        [0.0162, 0.7352, 0.7558],
        [0.0253, 0.7376, 0.7492],
        [0.0369, 0.74, 0.7426],
        [0.0504, 0.7423, 0.7359],
        [0.0638, 0.7446, 0.7292],
        [0.077, 0.7468, 0.7224],
        [0.0899, 0.7489, 0.7156],
        [0.1023, 0.751, 0.7088],
        [0.1141, 0.7531, 0.7019],
        [0.1252, 0.7552, 0.695],
        [0.1354, 0.7572, 0.6881],
        [0.1448, 0.7593, 0.6812],
        [0.1532, 0.7614, 0.6741],
        [0.1609, 0.7635, 0.6671],
        [0.1678, 0.7656, 0.6599],
        [0.1741, 0.7678, 0.6527],
        [0.1799, 0.7699, 0.6454],
        [0.1853, 0.7721, 0.6379],
        [0.1905, 0.7743, 0.6303],
        [0.1954, 0.7765, 0.6225],
        [0.2003, 0.7787, 0.6146],
        [0.2061, 0.7808, 0.6065],
        [0.2118, 0.7828, 0.5983],
        [0.2178, 0.7849, 0.5899],
        [0.2244, 0.7869, 0.5813],
        [0.2318, 0.7887, 0.5725],
        [0.2401, 0.7905, 0.5636],
        [0.2491, 0.7922, 0.5546],
        [0.2589, 0.7937, 0.5454],
        [0.2695, 0.7951, 0.536],
        [0.2809, 0.7964, 0.5266],
        [0.2929, 0.7975, 0.517],
        [0.3052, 0.7985, 0.5074],
        [0.3176, 0.7994, 0.4975],
        [0.3301, 0.8002, 0.4876],
        [0.3424, 0.8009, 0.4774],
        [0.3548, 0.8016, 0.4669],
        [0.3671, 0.8021, 0.4563],
        [0.3795, 0.8026, 0.4454],
        [0.3921, 0.8029, 0.4344],
        [0.405, 0.8031, 0.4233],
        [0.4184, 0.803, 0.4122],
        [0.4322, 0.8028, 0.4013],
        [0.4463, 0.8024, 0.3904],
        [0.4608, 0.8018, 0.3797],
        [0.4753, 0.8011, 0.3691],
        [0.4899, 0.8002, 0.3586],
        [0.5044, 0.7993, 0.348],
        [0.5187, 0.7982, 0.3374],
        [0.5329, 0.797, 0.3267],
        [0.547, 0.7957, 0.3159],
        [0.5609, 0.7943, 0.305],
        [0.5748, 0.7929, 0.2941],
        [0.5886, 0.7913, 0.2833],
        [0.6024, 0.7896, 0.2726],
        [0.6161, 0.7878, 0.2622],
        [0.6297, 0.7859, 0.2521],
        [0.6433, 0.7839, 0.2423],
        [0.6567, 0.7818, 0.2329],
        [0.6701, 0.7796, 0.2239],
        [0.6833, 0.7773, 0.2155],
        [0.6963, 0.775, 0.2075],
        [0.7091, 0.7727, 0.1998],
        [0.7218, 0.7703, 0.1924],
        [0.7344, 0.7679, 0.1852],
        [0.7468, 0.7654, 0.1782],
        [0.759, 0.7629, 0.1717],
        [0.771, 0.7604, 0.1658],
        [0.7829, 0.7579, 0.1608],
        [0.7945, 0.7554, 0.157],
        [0.806, 0.7529, 0.1546],
        [0.8172, 0.7505, 0.1535],
        [0.8281, 0.7481, 0.1536],
        [0.8389, 0.7457, 0.1546],
        [0.8495, 0.7435, 0.1564],
        [0.86, 0.7413, 0.1587],
        [0.8703, 0.7392, 0.1615],
        [0.8804, 0.7372, 0.165],
        [0.8903, 0.7353, 0.1695],
        [0.9, 0.7336, 0.1749],
        [0.9093, 0.7321, 0.1815],
        [0.9184, 0.7308, 0.189],
        [0.9272, 0.7298, 0.1973],
        [0.9357, 0.729, 0.2061],
        [0.944, 0.7285, 0.2151],
        [0.9523, 0.7284, 0.2237],
        [0.9606, 0.7285, 0.2312],
        [0.9689, 0.7292, 0.2373],
        [0.977, 0.7304, 0.2418],
        [0.9842, 0.733, 0.2446],
        [0.99, 0.7365, 0.2429],
        [0.9946, 0.7407, 0.2394],
        [0.9966, 0.7458, 0.2351],
        [0.9971, 0.7513, 0.2309],
        [0.9972, 0.7569, 0.2267],
        [0.9971, 0.7626, 0.2224],
        [0.9969, 0.7683, 0.2181],
        [0.9966, 0.774, 0.2138],
        [0.9962, 0.7798, 0.2095],
        [0.9957, 0.7856, 0.2053],
        [0.9949, 0.7915, 0.2012],
        [0.9938, 0.7974, 0.1974],
        [0.9923, 0.8034, 0.1939],
        [0.9906, 0.8095, 0.1906],
        [0.9885, 0.8156, 0.1875],
        [0.9861, 0.8218, 0.1846],
        [0.9835, 0.828, 0.1817],
        [0.9807, 0.8342, 0.1787],
        [0.9778, 0.8404, 0.1757],
        [0.9748, 0.8467, 0.1726],
        [0.972, 0.8529, 0.1695],
        [0.9694, 0.8591, 0.1665],
        [0.9671, 0.8654, 0.1636],
        [0.9651, 0.8716, 0.1608],
        [0.9634, 0.8778, 0.1582],
        [0.9619, 0.884, 0.1557],
        [0.9608, 0.8902, 0.1532],
        [0.9601, 0.8963, 0.1507],
        [0.9596, 0.9023, 0.148],
        [0.9595, 0.9084, 0.145],
        [0.9597, 0.9143, 0.1418],
        [0.9601, 0.9203, 0.1382],
        [0.9608, 0.9262, 0.1344],
        [0.9618, 0.932, 0.1304],
        [0.9629, 0.9379, 0.1261],
        [0.9642, 0.9437, 0.1216],
        [0.9657, 0.9494, 0.1168],
        [0.9674, 0.9552, 0.1116],
        [0.9692, 0.9609, 0.1061],
        [0.9711, 0.9667, 0.1001],
        [0.973, 0.9724, 0.0938],
        [0.9749, 0.9782, 0.0872],
        [0.9769, 0.9839, 0.0805]
    ])

    mesh_rec = trimesh.load(in_file_rec_meshes)
    mesh_gt = trimesh.load(in_file_gt_meshes)

    vertices_rec = mesh_rec.vertices

    closest_point_on_gt, vertices_rec_dist, triangle_ids = trimesh.proximity.closest_point(mesh_gt, vertices_rec)

    max_dist = np.max(vertices_rec_dist)
    vertices_rec_dist_normalized = vertices_rec_dist / max_dist

    print('Distance from reconstructed mesh vertex to nearest sample on GT mesh, Min={}, Max={}, Mean={}'.format(
        np.min(vertices_rec_dist), np.max(vertices_rec_dist), np.mean(vertices_rec_dist)
    ))

    if normalize_to is not None:
        if max_dist > normalize_to:
            print('test')
            raise ValueError('ERROR: given normalization maximum {} is less than the maximum distance here {}!'.format(
                normalize_to, max_dist
            ))
        vertices_rec_dist_normalized = vertices_rec_dist_normalized * (max_dist / normalize_to)

    # # dist=0 -> green, dist=1 -> red
    # # red is at 0 degree, green at 120 degree
    # vertices_rec_dist_hue = (1.0 - vertices_rec_dist_normalized) / 3.0
    # dist_colors_hsv = np.zeros((vertices_rec_dist_normalized.shape[0], 3))
    # dist_colors_hsv[:, 0] = vertices_rec_dist_hue
    # dist_colors_hsv[:, 1] = 1.0
    # dist_colors_hsv[:, 2] = 1.0
    # dist_colors_rgb = hsv2rgb(np.expand_dims(dist_colors_hsv, 1)).squeeze()

    # use parula colormap: dist=0 -> blue, dist=0.5 -> green, dist=1.0 -> yellow
    parula_indices = (vertices_rec_dist_normalized * (parula_cm.shape[0] - 1)).astype(np.int32)
    dist_colors_rgb = parula_cm[parula_indices]

    mesh_vis = trimesh.Trimesh(vertices=mesh_rec.vertices, faces=mesh_rec.faces, vertex_colors=dist_colors_rgb)
    mesh_vis.export(out_file_vis_meshes)


def _unit_test_comp_predictions():

    gt =   np.array([1, 1, 1, 0, 0, 0, -1, -1, -1])
    pred = np.array([1, 0, -1, 0, 1, -1, -1, 1, 0])
    #gt =   np.array([1, 1, 1, -1, -1, -1])
    #pred = np.array([1, -1, 1, -1, -1, -1])
    #gt =   np.array([1, 1, 1, 1, 1, 1])
    #pred = np.array([1, 1, 1, 1, 1, 1])
    #gt =   np.array([-1])
    #pred = np.array([-1])
    #gt =   np.array([-1])
    #pred = np.array([1])
    #geodesic_mat_gt = sparse.csr_matrix(gt_arr_test, dtype=int)
    #geodesic_mat_pred = sparse.csr_matrix(pr_arr_test, dtype=int)
    #geodesic_mat_gt.prune()
    #geodesic_mat_pred.prune()

    results = compare_predictions_multi_class(gt, pred, 'test')

    print_list_of_dicts(results)

    return 0
